@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -80,9 +80,9 @@ struct connman_element {
 	enum connman_element_state state;
 	enum connman_element_error error;
 	gboolean enabled;
-	gboolean configuring;
 	gchar *devname;
 
+	GHashTable *children;
 	struct connman_element *parent;
 
 	struct connman_driver *driver;
@@ -107,6 +107,7 @@ struct connman_element {
 		gchar *broadcast;
 		gchar *nameserver;
 		gchar *timeserver;
+		gchar *pac;
 	} ipv4;
 };
 
@@ -138,6 +139,8 @@ int connman_element_register(struct connman_element *element,
 					struct connman_element *parent);
 void connman_element_unregister(struct connman_element *element);
 void connman_element_unregister_children(struct connman_element *element);
+void connman_element_unregister_children_type(struct connman_element *element,
+						enum connman_element_type type);
 void connman_element_update(struct connman_element *element);
 
 int connman_element_set_enabled(struct connman_element *element,
