@@ -134,8 +134,6 @@ static const char *type2description(enum connman_device_type type)
 	case CONNMAN_DEVICE_TYPE_GPS:
 		return "GPS";
 	case CONNMAN_DEVICE_TYPE_CELLULAR:
-	case CONNMAN_DEVICE_TYPE_MBM:
-	case CONNMAN_DEVICE_TYPE_HSO:
 		return "Cellular";
 	}
 
@@ -159,8 +157,6 @@ static const char *type2string(enum connman_device_type type)
 	case CONNMAN_DEVICE_TYPE_GPS:
 		return "gps";
 	case CONNMAN_DEVICE_TYPE_CELLULAR:
-	case CONNMAN_DEVICE_TYPE_MBM:
-	case CONNMAN_DEVICE_TYPE_HSO:
 		return "cellular";
 	}
 
@@ -185,8 +181,6 @@ enum connman_service_type __connman_device_get_service_type(struct connman_devic
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 		return CONNMAN_SERVICE_TYPE_BLUETOOTH;
 	case CONNMAN_DEVICE_TYPE_CELLULAR:
-	case CONNMAN_DEVICE_TYPE_MBM:
-	case CONNMAN_DEVICE_TYPE_HSO:
 		return CONNMAN_SERVICE_TYPE_CELLULAR;
 	}
 
@@ -760,8 +754,6 @@ struct connman_device *connman_device_create(const char *node,
 		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_CELLULAR:
-	case CONNMAN_DEVICE_TYPE_MBM:
-	case CONNMAN_DEVICE_TYPE_HSO:
 		device->scan_interval = 0;
 		break;
 	}
@@ -973,6 +965,11 @@ int connman_device_set_powered(struct connman_device *device,
 	if (device->powered == powered)
 		return -EALREADY;
 
+	if (powered == TRUE)
+		__connman_device_enable(device);
+	else
+		__connman_device_disable(device);
+
 	device->powered = powered;
 	device->powered_pending = powered;
 
@@ -1187,24 +1184,6 @@ static void scanning_changed(struct connman_device *device)
 					DBUS_TYPE_BOOLEAN, &device->scanning);
 }
 
-static void mark_network_available(gpointer key, gpointer value,
-						gpointer user_data)
-{
-	struct connman_network *network = value;
-
-	connman_network_set_available(network, TRUE);
-}
-
-void connman_device_cleanup_scanning(struct connman_device *device)
-{
-	device->scanning = FALSE;
-
-	scanning_changed(device);
-
-	g_hash_table_foreach(device->networks,
-				mark_network_available, NULL);
-}
-
 /**
  * connman_device_set_scanning:
  * @device: device structure
@@ -1469,6 +1448,9 @@ void __connman_device_set_network(struct connman_device *device,
 {
 	const char *name;
 
+	if (device == NULL)
+		return;
+
 	if (device->network == network)
 		return;
 
